"""Helpers for cross compatibility across dependency versions."""
from importlib import resources
from typing import Callable, Iterable

from docutils.nodes import Element


def findall(node: Element) -> Callable[..., Iterable[Element]]:
    """Iterate through"""
    # findall replaces traverse in docutils v0.18
    # note a difference is that findall is an iterator
    return getattr(node, "findall", node.traverse)


# TODO: >= Python 3.9, only use `resources.files` and drop `resources.read_text`
def read_text(module: resources.Package, filename: str) -> str:
    if hasattr(resources, "files"):
        return resources.files(module).joinpath(filename).read_text()
    return resources.read_text(module, filename)



from typing import List, Optional, Tuple

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.util.docutils import ReferenceRole, SphinxDirective, SphinxRole

from sphinx_design.shared import SEMANTIC_COLORS, make_choice, text_align

ROLE_NAME_BADGE_PREFIX = "bdg"
ROLE_NAME_LINK_PREFIX = "bdg-link"
ROLE_NAME_REF_PREFIX = "bdg-ref"
DIRECTIVE_NAME_BUTTON_LINK = "button-link"
DIRECTIVE_NAME_BUTTON_REF = "button-ref"

# TODO defining arbitrary classes for badges
# (maybe split text right of last `;`, then split that by comma)
# in particular for rounded-pill class etc


def setup_badges_and_buttons(app: Sphinx) -> None:
    """Setup the badge components."""
    app.add_role(ROLE_NAME_BADGE_PREFIX, BadgeRole())
    app.add_role(ROLE_NAME_LINK_PREFIX, LinkBadgeRole())
    app.add_role(ROLE_NAME_REF_PREFIX, XRefBadgeRole())
    for color in SEMANTIC_COLORS:
        app.add_role("-".join((ROLE_NAME_BADGE_PREFIX, color)), BadgeRole(color))
        app.add_role(
            "-".join((ROLE_NAME_BADGE_PREFIX, color, "line")),
            BadgeRole(color, outline=True),
        )
        app.add_role("-".join((ROLE_NAME_LINK_PREFIX, color)), LinkBadgeRole(color))
        app.add_role(
            "-".join((ROLE_NAME_LINK_PREFIX, color, "line")),
            LinkBadgeRole(color, outline=True),
        )
        app.add_role("-".join((ROLE_NAME_REF_PREFIX, color)), XRefBadgeRole(color))
        app.add_role(
            "-".join((ROLE_NAME_REF_PREFIX, color, "line")),
            XRefBadgeRole(color, outline=True),
        )

    app.add_directive(DIRECTIVE_NAME_BUTTON_LINK, ButtonLinkDirective)
    app.add_directive(DIRECTIVE_NAME_BUTTON_REF, ButtonRefDirective)


def create_bdg_classes(color: Optional[str], outline: bool) -> List[str]:
    """Create the badge classes."""
    classes = [
        "sd-sphinx-override",
        "sd-badge",
    ]
    if color is None:
        return classes
    if outline:
        classes.extend([f"sd-outline-{color}", f"sd-text-{color}"])
    else:
        classes.extend([f"sd-bg-{color}", f"sd-bg-text-{color}"])
    return classes


class BadgeRole(SphinxRole):
    """Role to display a badge."""

    def __init__(self, color: Optional[str] = None, *, outline: bool = False) -> None:
        super().__init__()
        self.color = color
        self.outline = outline

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        node = nodes.inline(
            self.rawtext,
            self.text,
            classes=create_bdg_classes(self.color, self.outline),
        )
        self.set_source_info(node)
        return [node], []


class LinkBadgeRole(ReferenceRole):
    """Role to display a badge with an external link."""

    def __init__(self, color: Optional[str] = None, *, outline: bool = False) -> None:
        super().__init__()
        self.color = color
        self.outline = outline

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        node = nodes.reference(
            self.rawtext,
            refuri=self.target,
            classes=create_bdg_classes(self.color, self.outline),
        )
        # TODO open in new tab
        self.set_source_info(node)
        # if self.target != self.title:
        #     node["reftitle"] = self.target
        node += nodes.inline(self.title, self.title)
        return [node], []


class XRefBadgeRole(ReferenceRole):
    """Role to display a badge with an internal link."""

    def __init__(self, color: Optional[str] = None, *, outline: bool = False) -> None:
        super().__init__()
        self.color = color
        self.outline = outline

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        options = {
            "classes": create_bdg_classes(self.color, self.outline),
            "reftarget": self.target,
            "refdoc": self.env.docname,
            "refdomain": "",
            "reftype": "any",
            "refexplicit": self.has_explicit_title,
            "refwarn": True,
        }
        node = addnodes.pending_xref(self.rawtext, **options)
        self.set_source_info(node)
        node += nodes.inline(self.title, self.title, classes=["xref", "any"])
        return [node], []


class _ButtonDirective(SphinxDirective):
    """A base button directive."""

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "color": make_choice(SEMANTIC_COLORS),
        "outline": directives.flag,
        "align": text_align,
        # expand to fit parent width
        "expand": directives.flag,
        # make parent also clickable
        "click-parent": directives.flag,
        "tooltip": directives.unchanged_required,
        "shadow": directives.flag,
        # ref button only
        "ref-type": make_choice(["any", "ref", "doc", "myst"]),
        "class": directives.class_option,
    }

    def create_ref_node(
        self, rawtext: str, target: str, explicit_title: bool, classes: List[str]
    ) -> nodes.Node:
        """Create the reference node."""
        raise NotImplementedError

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        rawtext = self.arguments[0]
        target = directives.uri(rawtext)
        classes = ["sd-sphinx-override", "sd-btn", "sd-text-wrap"]
        if "color" in self.options:
            if "outline" in self.options:
                classes.append(f"sd-btn-outline-{self.options['color']}")
            else:
                classes.append(f"sd-btn-{self.options['color']}")
        if "click-parent" in self.options:
            classes.append("sd-stretched-link")
        if "shadow" in self.options:
            classes.append("sd-shadow-sm")
        if "class" in self.options:
            classes.extend(self.options["class"])
        node = self.create_ref_node(rawtext, target, bool(self.content), classes)
        # TODO open in new tab
        self.set_source_info(node)
        if "tooltip" in self.options:
            node["reftitle"] = self.options["tooltip"]  # TODO escape HTML

        if self.content:
            textnodes, _ = self.state.inline_text(
                "\n".join(self.content), self.lineno + self.content_offset
            )
            content = nodes.inline("", "")
            content.extend(textnodes)
        else:
            content = nodes.inline(target, target)
        node.append(content)

        if "expand" in self.options:
            grid_container = nodes.inline(classes=["sd-d-grid"])
            self.set_source_info(grid_container)
            grid_container += node
            node = grid_container

        # `visit_reference` requires that a reference be inside a `TextElement` parent
        container = nodes.paragraph(classes=self.options.get("align", []))
        self.set_source_info(container)
        container += node

        return [container]


class ButtonLinkDirective(_ButtonDirective):
    """A button directive with an external link."""

    def create_ref_node(
        self, rawtext: str, target: str, explicit_title: bool, classes: List[str]
    ) -> nodes.Node:
        """Create the reference node."""
        return nodes.reference(
            rawtext,
            refuri=target,
            classes=classes,
        )


class ButtonRefDirective(_ButtonDirective):
    """A button directive with an internal link."""

    def create_ref_node(
        self, rawtext: str, target: str, explicit_title: bool, classes: List[str]
    ) -> nodes.Node:
        """Create the reference node."""
        ref_type = self.options.get("ref-type", "any")
        options = {
            # TODO the presence of classes raises an error if the link cannot be found
            "classes": classes,
            "reftarget": target,
            "refdoc": self.env.docname,
            "refdomain": "std" if ref_type in {"ref", "doc"} else "",
            "reftype": ref_type,
            "refexplicit": explicit_title,
            "refwarn": True,
        }
        return addnodes.pending_xref(rawtext, **options)



import hashlib
from pathlib import Path

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.transforms import SphinxTransform
from sphinx.util.docutils import SphinxDirective

from . import compiled as static_module
from ._compat import findall, read_text
from .article_info import setup_article_info
from .badges_buttons import setup_badges_and_buttons
from .cards import setup_cards
from .dropdown import setup_dropdown
from .grids import setup_grids
from .icons import setup_icons
from .shared import PassthroughTextElement, create_component
from .tabs import setup_tabs


def setup_extension(app: Sphinx) -> None:
    """Set up the sphinx extension."""
    app.connect("builder-inited", update_css_js)
    app.connect("env-updated", update_css_links)
    # we override container html visitors, to stop the default behaviour
    # of adding the `container` class to all nodes.container
    app.add_node(
        nodes.container, override=True, html=(visit_container, depart_container)
    )
    app.add_node(
        PassthroughTextElement,
        html=(visit_depart_null, visit_depart_null),
        latex=(visit_depart_null, visit_depart_null),
        text=(visit_depart_null, visit_depart_null),
        man=(visit_depart_null, visit_depart_null),
        texinfo=(visit_depart_null, visit_depart_null),
    )
    app.add_directive(
        "div", Div, override=True
    )  # override sphinx-panels implementation
    app.add_transform(AddFirstTitleCss)
    setup_badges_and_buttons(app)
    setup_cards(app)
    setup_grids(app)
    setup_dropdown(app)
    setup_icons(app)
    setup_tabs(app)
    setup_article_info(app)


def update_css_js(app: Sphinx):
    """Copy the CSS to the build directory."""
    # reset changed identifier
    app.env.sphinx_design_css_changed = False
    # setup up new static path in output dir
    static_path = (Path(app.outdir) / "_sphinx_design_static").absolute()
    static_existed = static_path.exists()
    static_path.mkdir(exist_ok=True)
    app.config.html_static_path.append(str(static_path))
    # Copy JS to the build directory.
    js_path = static_path / "design-tabs.js"
    app.add_js_file(js_path.name)
    if not js_path.exists():
        content = read_text(static_module, "sd_tabs.js")
        js_path.write_text(content)
    # Read the css content and hash it
    content = read_text(static_module, "style.min.css")
    hash = hashlib.md5(content.encode("utf8")).hexdigest()
    # Write the css file
    css_path = static_path / f"design-style.{hash}.min.css"
    app.add_css_file(css_path.name)
    if css_path.exists():
        return
    if static_existed:
        app.env.sphinx_design_css_changed = True
    for path in static_path.glob("*.css"):
        path.unlink()
    css_path.write_text(content, encoding="utf8")


def update_css_links(app: Sphinx, env: BuildEnvironment):
    """If CSS has changed, all files must be re-written, to include the correct stylesheets."""
    if env.sphinx_design_css_changed:
        return list(env.all_docs.keys())


def visit_container(self, node: nodes.Node):
    classes = "docutils container"
    attrs = {}
    if node.get("is_div", False):
        # we don't want the CSS for container for these nodes
        classes = "docutils"
    if "style" in node:
        attrs["style"] = node["style"]
    self.body.append(self.starttag(node, "div", CLASS=classes, **attrs))


def depart_container(self, node: nodes.Node):
    self.body.append("</div>\n")


def visit_depart_null(self, node: nodes.Element) -> None:
    """visit/depart passthrough"""


class Div(SphinxDirective):
    """Same as the ``container`` directive,
    but does not add the ``container`` class in HTML outputs,
    which can interfere with Bootstrap CSS.
    """

    optional_arguments = 1  # css classes
    final_argument_whitespace = True
    option_spec = {"style": directives.unchanged, "name": directives.unchanged}
    has_content = True

    def run(self):
        try:
            if self.arguments:
                classes = directives.class_option(self.arguments[0])
            else:
                classes = []
        except ValueError:
            raise self.error(
                'Invalid class attribute value for "%s" directive: "%s".'
                % (self.name, self.arguments[0])
            )
        node = create_component("div", rawtext="\n".join(self.content), classes=classes)
        if "style" in self.options:
            node["style"] = self.options["style"]
        self.set_source_info(node)
        self.add_name(node)
        if self.content:
            self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


class AddFirstTitleCss(SphinxTransform):
    """Add a CSS class to to the first sections title."""

    default_priority = 699  # priority main

    def apply(self):
        hide = False
        for docinfo in findall(self.document)(nodes.docinfo):
            for name in findall(docinfo)(nodes.field_name):
                if name.astext() == "sd_hide_title":
                    hide = True
                    break
            break
        if not hide:
            return
        for section in findall(self.document)(nodes.section):
            if isinstance(section.children[0], nodes.title):
                if "classes" in section.children[0]:
                    section.children[0]["classes"].append("sd-d-none")
                else:
                    section.children[0]["classes"] = ["sd-d-none"]
            break



from typing import List, Optional

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective

from .icons import get_octicon
from .shared import SEMANTIC_COLORS, create_component, make_choice


def setup_article_info(app: Sphinx):
    """Setup the article information components."""
    app.add_directive("article-info", ArticleInfoDirective)


class ArticleInfoDirective(SphinxDirective):
    """ """

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    option_spec = {
        "avatar": directives.uri,
        "avatar-alt": directives.unchanged,
        "avatar-link": directives.uri,
        "avatar-outline": make_choice(SEMANTIC_COLORS),
        "author": directives.unchanged_required,
        "date": directives.unchanged_required,
        "read-time": directives.unchanged_required,
        "class-container": directives.class_option,
        "class-avatar": directives.class_option,
    }

    def _parse_text(
        self, text: str, icon: Optional[nodes.Node] = None, parse: bool = False
    ) -> nodes.Node:
        """Parse the text."""
        if not parse:
            output = ([icon] if icon else []) + [nodes.Text(text)]
        else:
            text_nodes, _ = self.state.inline_text(text, self.lineno)
            text_nodes = ([icon] if icon else []) + text_nodes
            # note certain nodes (like references) need to be nested in a TextElement node
            # (e.g. a pargraph)
            para = nodes.paragraph("", "", *text_nodes, classes=["sd-p-0", "sd-m-0"])
            self.set_source_info(para)
            output = [para]
        return output

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        parse_fields = True  # parse field text

        top_grid = create_component(
            "grid-container",
            [
                "sd-container-fluid",
                "sd-sphinx-override",
                "sd-p-0",
                "sd-mt-2",
                "sd-mb-4",
            ]
            + self.options.get("class-container", []),
        )
        self.set_source_info(top_grid)

        top_row = create_component(
            "grid-row",
            ["sd-row", "sd-row-cols-2", "sd-gx-2", "sd-gy-1"],
        )
        self.set_source_info(top_row)
        top_grid += top_row

        avatar_uri = self.options.get("avatar")
        if avatar_uri:
            # TODO only in html (hide in latex)
            avatar_column = create_component(
                "grid-item",
                ["sd-col", "sd-col-auto", "sd-d-flex-row", "sd-align-minor-center"],
            )
            self.set_source_info(avatar_column)
            avatar_classes = ["sd-avatar-sm"]
            if "avatar-outline" in self.options:
                avatar_classes.append(f"sd-outline-{self.options['avatar-outline']}")
            if "class-avatar" in self.options:
                avatar_classes += self.options["class-avatar"]
            avatar_image = nodes.image(
                "",
                uri=avatar_uri,
                alt=self.options.get("avatar-alt", ""),
                classes=avatar_classes,
            )
            self.set_source_info(avatar_image)
            if self.options.get("avatar-link"):
                avatar_link = nodes.reference(
                    "", "", refuri=self.options.get("avatar-link")
                )
                avatar_link += avatar_image
                avatar_image = avatar_link
            avatar_column += avatar_image
            top_row += avatar_column

        info_column = create_component(
            "grid-item",
            ["sd-col", "sd-d-flex-row", "sd-align-minor-center"],
        )
        self.set_source_info(info_column)
        top_row += info_column

        info_grid = create_component(
            "grid-container",
            [
                "sd-container-fluid",
                "sd-sphinx-override",
            ],
        )
        self.set_source_info(info_grid)
        info_column += info_grid

        info_row = create_component(
            "grid-row",
            [
                "sd-row",
                "sd-row-cols-2",
                "sd-row-cols-xs-2",
                "sd-row-cols-sm-3",
                "sd-row-cols-md-3",
                "sd-row-cols-lg-3",
                "sd-gx-3",
                "sd-gy-1",
            ],
        )
        self.set_source_info(info_row)
        info_grid += info_row

        author_text = self.options.get("author")
        if author_text:
            author_column = create_component(
                "grid-item",
                ["sd-col", "sd-col-auto", "sd-d-flex-row", "sd-align-minor-center"],
            )
            self.set_source_info(author_column)
            author_nodes = self._parse_text(author_text, parse=parse_fields)
            author_column.extend(author_nodes)
            info_row += author_column

        date_text = self.options.get("date")
        if date_text:
            date_column = create_component(
                "grid-item",
                ["sd-col", "sd-col-auto", "sd-d-flex-row", "sd-align-minor-center"],
            )
            self.set_source_info(date_column)
            date_icon = nodes.raw(
                "",
                nodes.Text(get_octicon("calendar", height="16px")),
                classes=["sd-pr-2"],
                format="html",
            )
            date_nodes = self._parse_text(date_text, icon=date_icon, parse=parse_fields)
            date_column.extend(date_nodes)
            info_row += date_column

        read_time_text = self.options.get("read-time")
        if read_time_text:
            read_time_column = create_component(
                "grid-item",
                ["sd-col", "sd-col-auto", "sd-d-flex-row", "sd-align-minor-center"],
            )
            self.set_source_info(read_time_column)
            read_time_icon = nodes.raw(
                "",
                nodes.Text(get_octicon("clock", height="16px")),
                classes=["sd-pr-2"],
                format="html",
            )
            read_time_nodes = self._parse_text(
                read_time_text, icon=read_time_icon, parse=parse_fields
            )
            read_time_column.extend(read_time_nodes)
            info_row += read_time_column

        return [top_grid]



"""A sphinx extension for designing beautiful, view size responsive web components."""
from typing import TYPE_CHECKING

__version__ = "0.5.0"

if TYPE_CHECKING:
    from sphinx.application import Sphinx


def setup(app: "Sphinx") -> dict:
    from .extension import setup_extension

    setup_extension(app)
    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Originally Adapted from sphinxcontrib.details.directive
"""
from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.docutils import SphinxDirective

from sphinx_design.shared import (
    SEMANTIC_COLORS,
    create_component,
    is_component,
    make_choice,
    margin_option,
)

from ._compat import findall
from .icons import get_octicon, list_octicons


def setup_dropdown(app: Sphinx) -> None:
    app.add_node(dropdown_main, html=(visit_dropdown_main, depart_dropdown_main))
    app.add_node(dropdown_title, html=(visit_dropdown_title, depart_dropdown_title))
    app.add_directive("dropdown", DropdownDirective)
    app.add_post_transform(DropdownHtmlTransform)


class dropdown_main(nodes.Element, nodes.General):
    pass


class dropdown_title(nodes.TextElement, nodes.General):
    pass


def visit_dropdown_main(self, node):
    if node.get("opened"):
        self.body.append(self.starttag(node, "details", open="open"))
    else:
        self.body.append(self.starttag(node, "details"))


def depart_dropdown_main(self, node):
    self.body.append("</details>")


def visit_dropdown_title(self, node):
    self.body.append(self.starttag(node, "summary"))


def depart_dropdown_title(self, node):
    self.body.append("</summary>")


class DropdownDirective(SphinxDirective):
    """A directive to generate a collapsible container.

    Note: This directive generates a single container,
    for the title (optional) and content::

        <container design_component="dropdown" has_title=True>
            <rubric>
                ...title nodes
        ...content nodes

    This allows for a default rendering in non-HTML outputs.

    The ``DropdownHtmlTransform`` then transforms this container
    into the HTML specific structure.
    """

    optional_arguments = 1  # title of dropdown
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "open": directives.flag,  # make open by default
        "color": make_choice(SEMANTIC_COLORS),
        "icon": make_choice(list_octicons()),
        "animate": make_choice(("fade-in", "fade-in-slide-down")),
        "margin": margin_option,
        "name": directives.unchanged,
        "class-container": directives.class_option,
        "class-title": directives.class_option,
        "class-body": directives.class_option,
    }

    def run(self):
        """Run the directive"""
        # default classes
        classes = {
            "container_classes": self.options.get("margin", ["sd-mb-3"])
            + self.options.get("class-container", []),
            "title_classes": self.options.get("class-title", []),
            "body_classes": self.options.get("class-body", []),
        }

        # add color classes
        title_color = self.options.get("color")
        if title_color:
            classes["title_classes"].extend(
                [f"sd-bg-{title_color}", f"sd-bg-text-{title_color}"]
            )

        # add animation classes
        if (
            "animate" in self.options
            and ("sd-" + self.options["animate"]) not in classes["container_classes"]
        ):
            classes["container_classes"].append("sd-" + self.options["animate"])

        container = create_component(
            "dropdown",
            opened="open" in self.options,
            type="dropdown",
            has_title=len(self.arguments) > 0,
            icon=self.options.get("icon"),
            **classes,
        )
        self.set_source_info(container)
        if self.arguments:
            textnodes, messages = self.state.inline_text(self.arguments[0], self.lineno)
            title_node = nodes.rubric(self.arguments[0], "", *textnodes)
            container += title_node
            container += messages
            # where possible we add the target to the title node,
            # so that it can be used as the reference text
            self.add_name(title_node)
        else:
            self.add_name(container)
        self.state.nested_parse(self.content, self.content_offset, container)
        return [container]


class DropdownHtmlTransform(SphinxPostTransform):
    """Transform dropdown containers into the HTML specific AST structures::

    <details class="sd-sphinx-override sd-dropdown sd-card">
        <summary class="sd-summary-title sd-card-header">
            ...title nodes
        <div class="sd-summary-content sd-card-body">
            ...content nodes

    """

    default_priority = 199
    formats = ("html",)

    def run(self):
        """Run the transform"""
        document: nodes.document = self.document
        for node in findall(document)(lambda node: is_component(node, "dropdown")):
            # TODO option to not have card css (but requires more formatting)
            use_card = True

            open_marker = create_component(
                "dropdown-open-marker",
                classes=["sd-summary-up"],
                children=[
                    nodes.raw(
                        "",
                        nodes.Text(get_octicon("chevron-up", height="1.5em")),
                        format="html",
                    )
                ],
            )
            closed_marker = create_component(
                "dropdown-closed-marker",
                classes=["sd-summary-down"],
                children=[
                    nodes.raw(
                        "",
                        nodes.Text(get_octicon("chevron-down", height="1.5em")),
                        format="html",
                    )
                ],
            )

            newnode = dropdown_main(
                opened=node["opened"],
                classes=["sd-sphinx-override", "sd-dropdown"]
                + (["sd-card"] if use_card else ["sd-d-flex-column"])
                + node["container_classes"],
            )

            if node["has_title"]:
                title_children = node[0].children
                if node[0].get("ids"):
                    newnode["ids"] += node[0]["ids"]
                body_children = node[1:]
            else:
                title_children = [
                    nodes.raw(
                        "...",
                        nodes.Text(get_octicon("kebab-horizontal", height="1.5em")),
                        format="html",
                    )
                ]
                body_children = node.children
            if node["icon"]:
                title_children.insert(
                    0,
                    nodes.raw(
                        "",
                        nodes.Text(get_octicon(node["icon"], height="1em")),
                        classes=["sd-summary-icon"],
                        format="html",
                    ),
                )

            newnode += dropdown_title(
                "",
                "",
                *title_children,
                closed_marker,
                open_marker,
                classes=["sd-summary-title"]
                + (["sd-card-header"] if use_card else [])
                + node["title_classes"],
            )
            body_node = create_component(
                "dropdown-body",
                classes=["sd-summary-content"]
                + (["sd-card-body"] if use_card else [])
                + node["body_classes"],
                children=body_children,
            )
            if use_card:
                for para in findall(body_node)(nodes.paragraph):
                    para["classes"] = ([] if "classes" in para else para["classes"]) + [
                        "sd-card-text"
                    ]
            newnode += body_node
            # newnode += open_marker
            node.replace_self(newnode)



import os
from pathlib import Path
from typing import Any, Dict, Optional

from docutils import nodes
import pytest
from sphinx.testing.path import path as sphinx_path
from sphinx.testing.util import SphinxTestApp

from sphinx_design._compat import findall

pytest_plugins = "sphinx.testing.fixtures"


class SphinxBuilder:
    def __init__(self, app: SphinxTestApp, src_path: Path):
        self.app = app
        self._src_path = src_path

    @property
    def src_path(self) -> Path:
        return self._src_path

    @property
    def out_path(self) -> Path:
        return Path(self.app.outdir)

    def build(self, assert_pass=True):
        self.app.build()
        if assert_pass:
            assert self.warnings == "", self.status
        return self

    @property
    def status(self):
        return self.app._status.getvalue()

    @property
    def warnings(self):
        return self.app._warning.getvalue()

    def get_doctree(
        self, docname: str, post_transforms: bool = False
    ) -> nodes.document:
        doctree = self.app.env.get_doctree(docname)
        if post_transforms:
            self.app.env.apply_post_transforms(doctree, docname)
        # make source path consistent for test comparisons
        for node in findall(doctree)(include_self=True):
            if not ("source" in node and node["source"]):
                continue
            node["source"] = Path(node["source"]).relative_to(self.src_path).as_posix()
            if node["source"].endswith(".rst"):
                node["source"] = node["source"][:-4]
            elif node["source"].endswith(".md"):
                node["source"] = node["source"][:-3]
        # remove mathjax classes added by myst parser
        if doctree.children and isinstance(doctree.children[0], nodes.section):
            doctree.children[0]["classes"] = []
        return doctree


@pytest.fixture()
def sphinx_builder(tmp_path: Path, make_app, monkeypatch):
    def _create_project(
        buildername: str = "html", conf_kwargs: Optional[Dict[str, Any]] = None
    ):
        src_path = tmp_path / "srcdir"
        src_path.mkdir()
        conf_kwargs = conf_kwargs or {
            "extensions": ["myst_parser", "sphinx_design"],
            "myst_enable_extensions": ["colon_fence"],
        }
        content = "\n".join(
            [f"{key} = {value!r}" for key, value in conf_kwargs.items()]
        )
        src_path.joinpath("conf.py").write_text(content, encoding="utf8")
        app = make_app(
            srcdir=sphinx_path(os.path.abspath(str(src_path))), buildername=buildername
        )
        return SphinxBuilder(app, src_path)

    yield _create_project



"""Shared constants and functions."""

from typing import List, Optional, Sequence

from docutils import nodes
from docutils.parsers.rst import directives

WARNING_TYPE = "design"

SEMANTIC_COLORS = (
    "primary",
    "secondary",
    "success",
    "info",
    "warning",
    "danger",
    "light",
    "muted",
    "dark",
    "white",
    "black",
)


def create_component(
    name: str,
    classes: Sequence[str] = (),
    *,
    rawtext: str = "",
    children: Sequence[nodes.Node] = (),
    **attributes,
) -> nodes.container:
    """Create a container node for a design component."""
    node = nodes.container(
        rawtext, is_div=True, design_component=name, classes=list(classes), **attributes
    )
    node.extend(children)
    return node


def is_component(node: nodes.Node, name: str):
    """Check if a node is a certain design component."""
    try:
        return node.get("design_component") == name
    except AttributeError:
        return False


def make_choice(choices: Sequence[str]):
    """Create a choice validator."""
    return lambda argument: directives.choice(argument, choices)


def _margin_or_padding_option(
    argument: Optional[str],
    class_prefix: str,
    allowed: Sequence[str],
) -> List[str]:
    """Validate the margin/padding is one (all) or four (top bottom left right) integers,
    between 0 and 5 or 'auto'.
    """
    if argument is None:
        raise ValueError("argument required but none supplied")
    values = argument.split()
    for value in values:
        if value not in allowed:
            raise ValueError(f"{value} is not in: {allowed}")
    if len(values) == 1:
        return [f"{class_prefix}-{values[0]}"]
    if len(values) == 4:
        return [
            f"{class_prefix}{side}-{value}"
            for side, value in zip(["t", "b", "l", "r"], values)
        ]
    raise ValueError(
        "argument must be one (all) or four (top bottom left right) integers"
    )


def margin_option(argument: Optional[str]) -> List[str]:
    """Validate the margin is one (all) or four (top bottom left right) integers,
    between 0 and 5 or 'auto'.
    """
    return _margin_or_padding_option(
        argument, "sd-m", ("auto", "0", "1", "2", "3", "4", "5")
    )


def padding_option(argument: Optional[str]) -> List[str]:
    """Validate the padding is one (all) or four (top bottom left right) integers,
    between 0 and 5.
    """
    return _margin_or_padding_option(argument, "sd-p", ("0", "1", "2", "3", "4", "5"))


def text_align(argument: Optional[str]) -> List[str]:
    """Validate the text align is left, right, center or justify."""
    value = directives.choice(argument, ["left", "right", "center", "justify"])
    return [f"sd-text-{value}"]


class PassthroughTextElement(nodes.TextElement):
    """A text element which will not render anything.

    This is required for reference node to render correctly outside of paragraphs.
    Since sphinx expects them to be within a ``TextElement``:
    https://github.com/sphinx-doc/sphinx/blob/068f802df90ea790f89319094e407c4d5f6c26ff/sphinx/writers/html5.py#L224
    """



from typing import List, Optional

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
from sphinx.util.logging import getLogger

from .cards import CardDirective
from .shared import (
    WARNING_TYPE,
    create_component,
    is_component,
    make_choice,
    margin_option,
    padding_option,
    text_align,
)

LOGGER = getLogger(__name__)


DIRECTIVE_NAME_GRID = "grid"
DIRECTIVE_NAME_GRID_ITEM = "grid-item"
DIRECTIVE_NAME_GRID_ITEM_CARD = "grid-item-card"


def setup_grids(app: Sphinx):
    """Setup the grid components."""
    app.add_directive(DIRECTIVE_NAME_GRID, GridDirective)
    app.add_directive(DIRECTIVE_NAME_GRID_ITEM, GridItemDirective)
    app.add_directive(DIRECTIVE_NAME_GRID_ITEM_CARD, GridItemCardDirective)


def _media_option(
    argument: Optional[str],
    prefix: str,
    *,
    allow_auto: bool = False,
    min_num: int = 1,
    max_num: int = 12,
) -> List[str]:
    """Validate the number of columns (out of 12).

    One or four integers (for "xs sm md lg") between 1 and 12.
    """
    validate_error_msg = (
        "argument must be 1 or 4 (xs sm md lg) values, and each value should be "
        f"{'either auto or ' if allow_auto else ''}an integer from {min_num} to {max_num}"
    )
    if argument is None:
        raise ValueError(validate_error_msg)
    values = argument.strip().split()
    if len(values) == 1:
        values = [values[0], values[0], values[0], values[0]]
    if len(values) != 4:
        raise ValueError(validate_error_msg)
    for value in values:
        if allow_auto and value == "auto":
            continue
        try:
            int_value = int(value)
        except Exception:
            raise ValueError(validate_error_msg)
        if not (min_num <= int_value <= max_num):
            raise ValueError(validate_error_msg)
    return [f"{prefix}{values[0]}"] + [
        f"{prefix}{size}-{value}"
        for size, value in zip(["xs", "sm", "md", "lg"], values)
    ]


def row_columns_option(argument: Optional[str]) -> List[str]:
    """Validate the number of columns (out of 12) a grid row will have.

    One or four integers (for "xs sm md lg") between 1 and 12  (or 'auto').
    """
    return _media_option(argument, "sd-row-cols-", allow_auto=True)


def item_columns_option(argument: Optional[str]) -> List[str]:
    """Validate the number of columns (out of 12) a grid-item will take up.

    One or four integers (for "xs sm md lg") between 1 and 12 (or 'auto').
    """
    return _media_option(argument, "sd-col-", allow_auto=True)


def gutter_option(argument: Optional[str]) -> List[str]:
    """Validate the gutter size between grid items.

    One or four integers (for "xs sm md lg") between 0 and 5.
    """
    return _media_option(argument, "sd-g-", min_num=0, max_num=5)


class GridDirective(SphinxDirective):
    """A grid component, which is a container for grid items (i.e. columns)."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1  # columns
    final_argument_whitespace = True
    option_spec = {
        "gutter": gutter_option,
        "margin": margin_option,
        "padding": padding_option,
        "outline": directives.flag,
        "reverse": directives.flag,
        "class-container": directives.class_option,
        "class-row": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        try:
            column_classes = (
                row_columns_option(self.arguments[0]) if self.arguments else []
            )
        except ValueError as exc:
            raise self.error(f"Invalid directive argument: {exc}")
        self.assert_has_content()
        # container-fluid is 100% width for all breakpoints,
        # rather than the fixed width of the breakpoint (like container)
        grid_classes = ["sd-container-fluid", "sd-sphinx-override"]
        container = create_component(
            "grid-container",
            grid_classes
            + self.options.get("margin", ["sd-mb-4"])
            + self.options.get("padding", [])
            + (["sd-border-1"] if "outline" in self.options else [])
            + self.options.get("class-container", []),
        )
        self.set_source_info(container)
        row = create_component(
            "grid-row",
            ["sd-row"]
            + column_classes
            + self.options.get("gutter", [])
            + (["sd-flex-row-reverse"] if "reverse" in self.options else [])
            + self.options.get("class-row", []),
        )
        self.set_source_info(row)
        container += row
        self.state.nested_parse(self.content, self.content_offset, row)
        # each item in a row should be a column
        for item in row.children:
            if not is_component(item, "grid-item"):
                LOGGER.warning(
                    f"All children of a 'grid-row' "
                    f"should be 'grid-item' [{WARNING_TYPE}.grid]",
                    location=item,
                    type=WARNING_TYPE,
                    subtype="grid",
                )
                break
        return [container]


class GridItemDirective(SphinxDirective):
    """An item within a grid row.

    Can "occupy" 1 to 12 columns.
    """

    has_content = True
    option_spec = {
        "columns": item_columns_option,
        "margin": margin_option,
        "padding": padding_option,
        "child-direction": make_choice(["column", "row"]),
        "child-align": make_choice(["start", "end", "center", "justify", "spaced"]),
        "outline": directives.flag,
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        if not is_component(self.state_machine.node, "grid-row"):
            LOGGER.warning(
                f"The parent of a 'grid-item' should be a 'grid-row' [{WARNING_TYPE}.grid]",
                location=(self.env.docname, self.lineno),
                type=WARNING_TYPE,
                subtype="grid",
            )
        column = create_component(
            "grid-item",
            [
                "sd-col",
                f"sd-d-flex-{self.options.get('child-direction', 'column')}",
            ]
            + self.options.get("columns", [])
            + self.options.get("margin", [])
            + self.options.get("padding", [])
            + (
                [f'sd-align-major-{self.options["child-align"]}']
                if "child-align" in self.options
                else []
            )
            + (["sd-border-1"] if "outline" in self.options else [])
            + self.options.get("class", []),
        )
        self.set_source_info(column)
        self.state.nested_parse(self.content, self.content_offset, column)
        return [column]


class GridItemCardDirective(SphinxDirective):
    """An item within a grid row, with an internal card."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1  # card title
    final_argument_whitespace = True
    option_spec = {
        "columns": item_columns_option,
        "margin": margin_option,
        "padding": padding_option,
        "class-item": directives.class_option,
        # The options below must be sync'ed with CardDirective.option_spec (minus margin).
        "width": make_choice(["auto", "25%", "50%", "75%", "100%"]),
        "text-align": text_align,
        "img-background": directives.uri,
        "img-top": directives.uri,
        "img-bottom": directives.uri,
        "img-alt": directives.unchanged,
        "link": directives.uri,
        "link-type": make_choice(["url", "any", "ref", "doc"]),
        "link-alt": directives.unchanged,
        "shadow": make_choice(["none", "sm", "md", "lg"]),
        "class-card": directives.class_option,
        "class-body": directives.class_option,
        "class-title": directives.class_option,
        "class-header": directives.class_option,
        "class-footer": directives.class_option,
        "class-img-top": directives.class_option,
        "class-img-bottom": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        if not is_component(self.state_machine.node, "grid-row"):
            LOGGER.warning(
                f"The parent of a 'grid-item' should be a 'grid-row' [{WARNING_TYPE}.grid]",
                location=(self.env.docname, self.lineno),
                type=WARNING_TYPE,
                subtype="grid",
            )
        column = create_component(
            "grid-item",
            [
                "sd-col",
                "sd-d-flex-row",
            ]
            + self.options.get("columns", [])
            + self.options.get("margin", [])
            + self.options.get("padding", [])
            + self.options.get("class-item", []),
        )
        card_options = {
            key: value
            for key, value in self.options.items()
            if key
            in [
                "width",
                "text-align",
                "img-background",
                "img-top",
                "img-bottom",
                "img-alt",
                "link",
                "link-type",
                "link-alt",
                "shadow",
                "class-card",
                "class-body",
                "class-title",
                "class-header",
                "class-footer",
                "class-img-top",
                "class-img-bottom",
            ]
        }
        if "width" not in card_options:
            card_options["width"] = "100%"
        card_options["margin"] = []
        card = CardDirective.create_card(self, self.arguments, card_options)
        column += card
        return [column]



from functools import lru_cache
import json
import re
from typing import Any, Dict, List, Optional, Sequence, Tuple

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.docutils import SphinxDirective, SphinxRole

from . import compiled
from ._compat import read_text
from .shared import WARNING_TYPE

logger = logging.getLogger(__name__)

OCTICON_VERSION = "v16.1.1"

OCTICON_CSS = """\
.octicon {
  display: inline-block;
  vertical-align: text-top;
  fill: currentColor;
}"""


def setup_icons(app: Sphinx) -> None:
    app.add_role("octicon", OcticonRole())
    app.add_directive("_all-octicon", AllOcticons)
    for style in ["fa", "fas", "fab", "far"]:
        # note: fa is deprecated in v5, fas is the default and fab is the other free option
        app.add_role(style, FontawesomeRole(style))
    for style in ["regular", "outlined", "round", "sharp", "twotone"]:
        app.add_role("material-" + style, MaterialRole(style))
    app.add_config_value("sd_fontawesome_latex", False, "env")
    app.connect("config-inited", add_fontawesome_pkg)
    app.add_node(
        fontawesome,
        html=(visit_fontawesome_html, depart_fontawesome_html),
        latex=(visit_fontawesome_latex, None),
        man=(visit_fontawesome_warning, None),
        text=(visit_fontawesome_warning, None),
        texinfo=(visit_fontawesome_warning, None),
    )


@lru_cache(1)
def get_octicon_data() -> Dict[str, Any]:
    """Load all octicon data."""
    content = read_text(compiled, "octicons.json")
    return json.loads(content)


def list_octicons() -> List[str]:
    """List available octicon names."""
    return list(get_octicon_data().keys())


HEIGHT_REGEX = re.compile(r"^(?P<value>\d+(\.\d+)?)(?P<unit>px|em|rem)$")


def get_octicon(
    name: str,
    height: str = "1em",
    classes: Sequence[str] = (),
    aria_label: Optional[str] = None,
) -> str:
    """Return the HTML for an GitHub octicon SVG icon.

    :height: the height of the octicon, with suffix unit 'px', 'em' or 'rem'.
    """
    try:
        data = get_octicon_data()[name]
    except KeyError:
        raise KeyError(f"Unrecognised octicon: {name}")

    match = HEIGHT_REGEX.match(height)
    if not match:
        raise ValueError(
            f"Invalid height: '{height}', must be format <integer><px|em|rem>"
        )
    height_value = round(float(match.group("value")), 3)
    height_unit = match.group("unit")

    original_height = 16
    if "16" not in data["heights"]:
        original_height = int(list(data["heights"].keys())[0])
    elif "24" in data["heights"]:
        if height_unit == "px":
            if height_value >= 24:
                original_height = 24
        elif height_value >= 1.5:
            original_height = 24
    original_width = data["heights"][str(original_height)]["width"]
    width_value = round(original_width * height_value / original_height, 3)
    content = data["heights"][str(original_height)]["path"]
    options = {
        "version": "1.1",
        "width": f"{width_value}{height_unit}",
        "height": f"{height_value}{height_unit}",
        "class": " ".join(("sd-octicon", f"sd-octicon-{name}", *classes)),
    }

    options["viewBox"] = f"0 0 {original_width} {original_height}"

    if aria_label is not None:
        options["aria-label"] = aria_label
        options["role"] = "img"
    else:
        options["aria-hidden"] = "true"

    opt_string = " ".join(f'{k}="{v}"' for k, v in options.items())
    return f"<svg {opt_string}>{content}</svg>"


class OcticonRole(SphinxRole):
    """Role to display a GitHub octicon SVG.

    Additional classes can be added to the element after a semicolon.
    """

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        values = self.text.split(";") if ";" in self.text else [self.text]
        icon = values[0]
        height = "1em" if len(values) < 2 else values[1]
        classes = "" if len(values) < 3 else values[2]
        icon = icon.strip()
        try:
            svg = get_octicon(icon, height=height, classes=classes.split())
        except Exception as exc:
            msg = self.inliner.reporter.error(
                f"Invalid octicon content: {exc}",
                line=self.lineno,
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)
            return [prb], [msg]
        node = nodes.raw("", nodes.Text(svg), format="html")
        self.set_source_info(node)
        return [node], []


class AllOcticons(SphinxDirective):
    """Directive to generate all octicon icons.

    Primarily for self documentation.
    """

    option_spec = {
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        classes = self.options.get("class", [])
        list_node = nodes.bullet_list()
        for icon in list_octicons():
            item_node = nodes.list_item()
            item_node.extend(
                (
                    nodes.literal(icon, icon),
                    nodes.Text(": "),
                    nodes.raw(
                        "",
                        nodes.Text(get_octicon(icon, classes=classes)),
                        format="html",
                    ),
                )
            )
            list_node += item_node
        return [list_node]


class fontawesome(nodes.Element, nodes.General):
    """Node for rendering fontawesome icon."""


class FontawesomeRole(SphinxRole):
    """Role to display a Fontawesome icon.

    Additional classes can be added to the element after a semicolon.
    """

    def __init__(self, style: str) -> None:
        super().__init__()
        self.style = style

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        icon, classes = self.text.split(";", 1) if ";" in self.text else [self.text, ""]
        icon = icon.strip()
        node = fontawesome(
            icon=icon, classes=[self.style, f"fa-{icon}"] + classes.split()
        )
        self.set_source_info(node)
        return [node], []


def visit_fontawesome_html(self, node):
    self.body.append(self.starttag(node, "span", ""))


def depart_fontawesome_html(self, node):
    self.body.append("</span>")


def add_fontawesome_pkg(app, config):
    if app.config.sd_fontawesome_latex:
        app.add_latex_package("fontawesome")


def visit_fontawesome_latex(self, node):
    """Add latex fonteawesome icon, if configured, else warn."""
    if self.config.sd_fontawesome_latex:
        self.body.append(f"\\faicon{{{node['icon']}}}")
    else:
        logger.warning(
            "Fontawesome icons not included in LaTeX output, "
            f"consider 'sd_fontawesome_latex=True' [{WARNING_TYPE}.fa-build]",
            location=node,
            type=WARNING_TYPE,
            subtype="fa-build",
        )
    raise nodes.SkipNode


def visit_fontawesome_warning(self, node: nodes.Element) -> None:
    """Warn that fontawesome is not supported for this builder."""
    logger.warning(
        "Fontawesome icons not supported for builder: "
        f"{self.builder.name} [{WARNING_TYPE}.fa-build]",
        location=node,
        type=WARNING_TYPE,
        subtype="fa-build",
    )
    raise nodes.SkipNode


@lru_cache(1)
def get_material_icon_data(style: str) -> Dict[str, Any]:
    """Load all octicon data."""
    content = read_text(compiled, f"material_{style}.json")
    return json.loads(content)


def get_material_icon(
    style: str,
    name: str,
    height: str = "1em",
    classes: Sequence[str] = (),
    aria_label: Optional[str] = None,
) -> str:
    """Return the HTML for an Google material icon SVG icon.

    :height: the height of the material icon, with suffix unit 'px', 'em' or 'rem'.
    """
    try:
        data = get_material_icon_data(style)[name]
    except KeyError:
        raise KeyError(f"Unrecognised material-{style} icon: {name}")

    match = HEIGHT_REGEX.match(height)
    if not match:
        raise ValueError(
            f"Invalid height: '{height}', must be format <integer><px|em|rem>"
        )
    height_value = round(float(match.group("value")), 3)
    height_unit = match.group("unit")

    original_height = 20
    if "20" not in data["heights"]:
        original_height = int(list(data["heights"].keys())[0])
    elif "24" in data["heights"]:
        if height_unit == "px":
            if height_value >= 24:
                original_height = 24
        elif height_value >= 1.5:
            original_height = 24
    original_width = data["heights"][str(original_height)]["width"]
    width_value = round(original_width * height_value / original_height, 3)
    content = data["heights"][str(original_height)]["path"]
    options = {
        "version": "4.0.0.63c5cb3",
        "width": f"{width_value}{height_unit}",
        "height": f"{height_value}{height_unit}",
        "class": " ".join(("sd-material-icon", f"sd-material-icon-{name}", *classes)),
    }

    options["viewBox"] = f"0 0 {original_width} {original_height}"

    if aria_label is not None:
        options["aria-label"] = aria_label
        options["role"] = "img"
    else:
        options["aria-hidden"] = "true"

    opt_string = " ".join(f'{k}="{v}"' for k, v in options.items())
    return f"<svg {opt_string}>{content}</svg>"


class MaterialRole(SphinxRole):
    """Role to display a Material-* icon.

    Additional classes can be added to the element after a semicolon.
    """

    def __init__(self, style: str) -> None:
        super().__init__()
        self.style = style

    def run(self) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
        """Run the role."""
        values = self.text.split(";") if ";" in self.text else [self.text]
        icon = values[0]
        height = "1em" if len(values) < 2 else values[1]
        classes = "" if len(values) < 3 else values[2]
        icon = icon.strip()
        try:
            svg = get_material_icon(
                self.style, icon, height=height, classes=classes.split()
            )
        except Exception as exc:
            msg = self.inliner.reporter.error(
                f"Invalid material-{self.style} icon content: {type(exc)} {exc}",
                line=self.lineno,
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)
            return [prb], [msg]
        node = nodes.raw("", nodes.Text(svg), format="html")
        self.set_source_info(node)
        return [node], []



import re
from typing import List, NamedTuple, Optional, Tuple

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.statemachine import StringList
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxDirective
from sphinx.util.logging import getLogger

from ._compat import findall
from .shared import (
    WARNING_TYPE,
    PassthroughTextElement,
    create_component,
    is_component,
    make_choice,
    margin_option,
    text_align,
)

LOGGER = getLogger(__name__)

DIRECTIVE_NAME_CARD = "card"
DIRECTIVE_NAME_CAROUSEL = "card-carousel"
REGEX_HEADER = re.compile(r"^\^{3,}\s*$")
REGEX_FOOTER = re.compile(r"^\+{3,}\s*$")


def setup_cards(app: Sphinx) -> None:
    """Setup the card components."""
    app.add_directive(DIRECTIVE_NAME_CARD, CardDirective)
    app.add_directive(DIRECTIVE_NAME_CAROUSEL, CardCarouselDirective)


class CardContent(NamedTuple):
    """Split card into header (optional), body, footer (optional).

    (offset, content)
    """

    body: Tuple[int, StringList]
    header: Optional[Tuple[int, StringList]] = None
    footer: Optional[Tuple[int, StringList]] = None


class CardDirective(SphinxDirective):
    """A card component."""

    has_content = True
    required_arguments = 0
    optional_arguments = 1  # card title
    final_argument_whitespace = True
    option_spec = {
        "width": make_choice(["auto", "25%", "50%", "75%", "100%"]),
        "margin": margin_option,
        "text-align": text_align,
        "img-top": directives.uri,
        "img-bottom": directives.uri,
        "img-background": directives.uri,
        "img-alt": directives.unchanged,
        "link": directives.uri,
        "link-type": make_choice(["url", "any", "ref", "doc"]),
        "link-alt": directives.unchanged,
        "shadow": make_choice(["none", "sm", "md", "lg"]),
        "class-card": directives.class_option,
        "class-header": directives.class_option,
        "class-body": directives.class_option,
        "class-title": directives.class_option,
        "class-footer": directives.class_option,
        "class-img-top": directives.class_option,
        "class-img-bottom": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        return [self.create_card(self, self.arguments, self.options)]

    @classmethod
    def create_card(
        cls, inst: SphinxDirective, arguments: Optional[list], options: dict
    ) -> nodes.Node:
        """Run the directive."""
        # TODO better degradation for latex
        card_classes = ["sd-card", "sd-sphinx-override"]
        if "width" in options:
            card_classes += [f'sd-w-{options["width"].rstrip("%")}']
        card_classes += options.get("margin", ["sd-mb-3"])
        card_classes += [f"sd-shadow-{options.get('shadow', 'sm')}"]
        if "link" in options:
            card_classes += ["sd-card-hover"]
        card = create_component(
            "card",
            card_classes
            + options.get("text-align", [])
            + options.get("class-card", []),
        )
        inst.set_source_info(card)

        img_alt = options.get("img-alt") or ""

        container = card
        if "img-background" in options:
            card.append(
                nodes.image(
                    uri=options["img-background"],
                    classes=["sd-card-img"],
                    alt=img_alt,
                )
            )
            overlay = create_component("card-overlay", ["sd-card-img-overlay"])
            inst.set_source_info(overlay)
            card += overlay
            container = overlay

        if "img-top" in options:
            image_top = nodes.image(
                "",
                uri=options["img-top"],
                alt=img_alt,
                classes=["sd-card-img-top"] + options.get("class-img-top", []),
            )
            container.append(image_top)

        components = cls.split_content(inst.content, inst.content_offset)

        if components.header:
            container.append(
                cls._create_component(
                    inst, "header", options, components.header[0], components.header[1]
                )
            )

        body = cls._create_component(
            inst, "body", options, components.body[0], components.body[1]
        )
        if arguments:
            title = create_component(
                "card-title",
                ["sd-card-title", "sd-font-weight-bold"]
                + options.get("class-title", []),
            )
            textnodes, _ = inst.state.inline_text(arguments[0], inst.lineno)
            title_container = PassthroughTextElement()
            title_container.extend(textnodes)
            inst.set_source_info(title_container)
            title.append(title_container)
            body.insert(0, title)
        container.append(body)

        if components.footer:
            container.append(
                cls._create_component(
                    inst, "footer", options, components.footer[0], components.footer[1]
                )
            )

        if "img-bottom" in options:
            image_bottom = nodes.image(
                "",
                uri=options["img-bottom"],
                alt=img_alt,
                classes=["sd-card-img-bottom"] + options.get("class-img-bottom", []),
            )
            container.append(image_bottom)

        if "link" in options:
            link_container = PassthroughTextElement()
            _classes = ["sd-stretched-link"]
            _rawtext = options.get("link-alt") or ""
            if options.get("link-alt"):
                _classes.append("sd-hide-link-text")
            if options.get("link-type", "url") == "url":
                link = nodes.reference(
                    _rawtext,
                    "",
                    refuri=options["link"],
                    classes=_classes,
                )
                if options.get("link-alt"):
                    link.append(nodes.inline(_rawtext, _rawtext))
            else:
                options = {
                    # TODO the presence of classes raises an error if the link cannot be found
                    "classes": _classes,
                    "reftarget": options["link"],
                    "refdoc": inst.env.docname,
                    "refdomain": "" if options["link-type"] == "any" else "std",
                    "reftype": options["link-type"],
                    "refexplicit": True,
                    "refwarn": True,
                }
                link = addnodes.pending_xref(
                    _rawtext, nodes.inline(_rawtext, _rawtext), **options
                )
            inst.set_source_info(link)
            link_container += link
            container.append(link_container)

        return card

    @staticmethod
    def split_content(content: StringList, offset: int) -> CardContent:
        """Split the content into header, body and footer."""
        header_index, footer_index, header, footer = None, None, None, None
        body_offset = offset
        for index, line in enumerate(content):
            # match the first occurrence of a header regex
            if (header_index is None) and REGEX_HEADER.match(line):
                header_index = index
            # match the final occurrence of a footer regex
            if REGEX_FOOTER.match(line):
                footer_index = index
        if header_index is not None:
            header = (offset, content[:header_index])
            body_offset += header_index + 1
        if footer_index is not None:
            footer = (offset + footer_index + 1, content[footer_index + 1 :])
        body = (
            body_offset,
            content[
                (header_index + 1 if header_index is not None else None) : footer_index
            ],
        )
        return CardContent(body, header, footer)

    @classmethod
    def _create_component(
        cls,
        inst: SphinxDirective,
        name: str,
        options: dict,
        offset: int,
        content: StringList,
    ) -> nodes.container:
        """Create the header, body, or footer."""
        component = create_component(
            f"card-{name}", [f"sd-card-{name}"] + options.get(f"class-{name}", [])
        )
        inst.set_source_info(component)  # TODO set proper lines
        inst.state.nested_parse(content, offset, component)
        cls.add_card_child_classes(component)
        return component

    @staticmethod
    def add_card_child_classes(node):
        """Add classes to specific child nodes."""
        for para in findall(node)(nodes.paragraph):
            para["classes"] = ([] if "classes" not in para else para["classes"]) + [
                "sd-card-text"
            ]
        # for title in findall(node)(nodes.title):
        #     title["classes"] = ([] if "classes" not in title else title["classes"]) + [
        #         "sd-card-title"
        #     ]


class CardCarouselDirective(SphinxDirective):
    """A component, which is a container for cards in a single scrollable row."""

    has_content = True
    required_arguments = 1  # columns
    optional_arguments = 0
    option_spec = {
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        try:
            cols = make_choice([str(i) for i in range(1, 13)])(
                self.arguments[0].strip()
            )
        except ValueError as exc:
            raise self.error(f"Invalid directive argument: {exc}")
        container = create_component(
            "card-carousel",
            ["sd-sphinx-override", "sd-cards-carousel", f"sd-card-cols-{cols}"]
            + self.options.get("class", []),
        )
        self.set_source_info(container)
        self.state.nested_parse(self.content, self.content_offset, container)
        for item in container.children:
            if not is_component(item, "card"):
                LOGGER.warning(
                    "All children of a 'card-carousel' "
                    f"should be 'card' [{WARNING_TYPE}.card]",
                    location=item,
                    type=WARNING_TYPE,
                    subtype="card",
                )
                break
        return [container]



"""Test the documented snippets run correctly, and are the same for both RST and MyST."""
from pathlib import Path
from typing import Callable

import pytest

from .conftest import SphinxBuilder

SNIPPETS_PATH = Path(__file__).parent.parent / "docs" / "snippets"
SNIPPETS_GLOB_RST = list((SNIPPETS_PATH / "rst").glob("[!_]*"))
SNIPPETS_GLOB_MYST = list((SNIPPETS_PATH / "myst").glob("[!_]*"))


def write_assets(src_path: Path):
    """Write additional assets to the src directory."""
    src_path.joinpath("snippet.py").write_text("a = 1")
    src_path.joinpath("images").mkdir()
    src_path.joinpath("images", "ebp-logo.png").touch()
    src_path.joinpath("images", "particle_background.jpg").touch()


@pytest.mark.parametrize(
    "path",
    SNIPPETS_GLOB_RST,
    ids=[path.name[: -len(path.suffix)] for path in SNIPPETS_GLOB_RST],
)
def test_snippets_rst(
    sphinx_builder: Callable[..., SphinxBuilder], path: Path, file_regression
):
    """Test snippets written in RestructuredText (before post-transforms)."""
    builder = sphinx_builder()
    content = "Heading\n-------" + "\n\n" + path.read_text(encoding="utf8")
    builder.src_path.joinpath("index.rst").write_text(content, encoding="utf8")
    write_assets(builder.src_path)
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=False)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename=f"snippet_pre_{path.name[:-len(path.suffix)]}",
        extension=".xml",
        encoding="utf8",
    )


@pytest.mark.parametrize(
    "path",
    SNIPPETS_GLOB_MYST,
    ids=[path.name[: -len(path.suffix)] for path in SNIPPETS_GLOB_MYST],
)
def test_snippets_myst(
    sphinx_builder: Callable[..., SphinxBuilder], path: Path, file_regression
):
    """Test snippets written in MyST Markdown (before post-transforms)."""
    builder = sphinx_builder()
    content = "# Heading" + "\n\n\n" + path.read_text(encoding="utf8")
    builder.src_path.joinpath("index.md").write_text(content, encoding="utf8")
    write_assets(builder.src_path)
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=False)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename=f"snippet_pre_{path.name[:-len(path.suffix)]}",
        extension=".xml",
        encoding="utf8",
    )


@pytest.mark.parametrize(
    "path",
    SNIPPETS_GLOB_RST,
    ids=[path.name[: -len(path.suffix)] for path in SNIPPETS_GLOB_RST],
)
def test_snippets_rst_post(
    sphinx_builder: Callable[..., SphinxBuilder], path: Path, file_regression
):
    """Test snippets written in RestructuredText (after HTML post-transforms)."""
    builder = sphinx_builder()
    content = "Heading\n-------" + "\n\n" + path.read_text(encoding="utf8")
    builder.src_path.joinpath("index.rst").write_text(content, encoding="utf8")
    write_assets(builder.src_path)
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=True)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename=f"snippet_post_{path.name[:-len(path.suffix)]}",
        extension=".xml",
        encoding="utf8",
    )


@pytest.mark.parametrize(
    "path",
    SNIPPETS_GLOB_MYST,
    ids=[path.name[: -len(path.suffix)] for path in SNIPPETS_GLOB_MYST],
)
def test_snippets_myst_post(
    sphinx_builder: Callable[..., SphinxBuilder], path: Path, file_regression
):
    """Test snippets written in MyST Markdown (after HTML post-transforms)."""
    builder = sphinx_builder()
    content = "# Heading" + "\n\n\n" + path.read_text(encoding="utf8")
    builder.src_path.joinpath("index.md").write_text(content, encoding="utf8")
    write_assets(builder.src_path)
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=True)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename=f"snippet_post_{path.name[:-len(path.suffix)]}",
        extension=".xml",
        encoding="utf8",
    )


def test_sd_hide_title_rst(
    sphinx_builder: Callable[..., SphinxBuilder], file_regression
):
    """Test that the root title is hidden."""
    builder = sphinx_builder()
    content = ":sd_hide_title:\n\nHeading\n-------\n\ncontent"
    builder.src_path.joinpath("index.rst").write_text(content, encoding="utf8")
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=False)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename="sd_hide_title",
        extension=".xml",
        encoding="utf8",
    )


def test_sd_hide_title_myst(
    sphinx_builder: Callable[..., SphinxBuilder], file_regression
):
    """Test that the root title is hidden."""
    builder = sphinx_builder()
    content = "---\nsd_hide_title: true\n---\n\n# Heading\n\ncontent"
    builder.src_path.joinpath("index.md").write_text(content, encoding="utf8")
    builder.build()
    doctree = builder.get_doctree("index", post_transforms=False)
    doctree.attributes.pop("translation_progress", None)  # added in sphinx 7.1
    file_regression.check(
        doctree.pformat(),
        basename="sd_hide_title",
        extension=".xml",
        encoding="utf8",
    )



from typing import List

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.application import Sphinx
from sphinx.transforms.post_transforms import SphinxPostTransform
from sphinx.util.docutils import SphinxDirective
from sphinx.util.logging import getLogger

from ._compat import findall
from .shared import WARNING_TYPE, create_component, is_component

LOGGER = getLogger(__name__)


def setup_tabs(app: Sphinx) -> None:
    app.add_directive("tab-set", TabSetDirective)
    app.add_directive("tab-item", TabItemDirective)
    app.add_directive("tab-set-code", TabSetCodeDirective)
    app.add_post_transform(TabSetHtmlTransform)
    app.add_node(sd_tab_input, html=(visit_tab_input, depart_tab_input))
    app.add_node(sd_tab_label, html=(visit_tab_label, depart_tab_label))


class TabSetDirective(SphinxDirective):
    """A container for a set of tab items."""

    has_content = True
    option_spec = {
        "class": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        tab_set = create_component(
            "tab-set", classes=["sd-tab-set"] + self.options.get("class", [])
        )
        self.set_source_info(tab_set)
        self.state.nested_parse(self.content, self.content_offset, tab_set)
        for item in tab_set.children:
            if not is_component(item, "tab-item"):
                LOGGER.warning(
                    f"All children of a 'tab-set' "
                    f"should be 'tab-item' [{WARNING_TYPE}.tab]",
                    location=item,
                    type=WARNING_TYPE,
                    subtype="tab",
                )
                break
        return [tab_set]


class TabItemDirective(SphinxDirective):
    """A single tab item in a tab set.

    Note: This directive generates a single container,
    for the label and content::

        <container design_component="tab-item" has_title=True>
            <rubric>
                ...title nodes
            <container design_component="tab-content">
                ...content nodes

    This allows for a default rendering in non-HTML outputs.

    The ``TabHtmlTransform`` then transforms this container
    into the HTML specific structure.
    """

    required_arguments = 1  # the tab label is the first argument
    final_argument_whitespace = True
    has_content = True
    option_spec = {
        "selected": directives.flag,
        "sync": directives.unchanged_required,
        "name": directives.unchanged,
        "class-container": directives.class_option,
        "class-label": directives.class_option,
        "class-content": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        if not is_component(self.state_machine.node, "tab-set"):
            LOGGER.warning(
                f"The parent of a 'tab-item' should be a 'tab-set' [{WARNING_TYPE}.tab]",
                location=(self.env.docname, self.lineno),
                type=WARNING_TYPE,
                subtype="tab",
            )
        tab_item = create_component(
            "tab-item",
            classes=["sd-tab-item"] + self.options.get("class-container", []),
            selected=("selected" in self.options),
        )

        # add tab label
        textnodes, _ = self.state.inline_text(self.arguments[0], self.lineno)
        tab_label = nodes.rubric(
            self.arguments[0],
            *textnodes,
            classes=["sd-tab-label"] + self.options.get("class-label", []),
        )
        if "sync" in self.options:
            tab_label["sync_id"] = self.options["sync"]
        self.add_name(tab_label)
        tab_item += tab_label

        # add tab content
        tab_content = create_component(
            "tab-content",
            classes=["sd-tab-content"] + self.options.get("class-content", []),
        )
        self.state.nested_parse(self.content, self.content_offset, tab_content)
        tab_item += tab_content

        return [tab_item]


class TabSetCodeDirective(SphinxDirective):
    """A container for a set of tab items, generated from code blocks."""

    has_content = True
    option_spec = {
        "no-sync": directives.flag,
        "class-set": directives.class_option,
        "class-item": directives.class_option,
    }

    def run(self) -> List[nodes.Node]:
        """Run the directive."""
        self.assert_has_content()
        tab_set = create_component(
            "tab-set", classes=["sd-tab-set"] + self.options.get("class-set", [])
        )
        self.set_source_info(tab_set)
        self.state.nested_parse(self.content, self.content_offset, tab_set)
        new_children = []
        for item in tab_set.children:
            if not isinstance(item, nodes.literal_block):
                LOGGER.warning(
                    f"All children of a 'tab-code-set' "
                    f"should be a 'literal_block' [{WARNING_TYPE}.tab_code]",
                    location=tab_set,
                    type=WARNING_TYPE,
                    subtype="tab_code",
                )
                continue
            language = item.get("language", "unknown")
            tab_label = nodes.rubric(
                language.upper(),
                nodes.Text(language.upper()),
                classes=["sd-tab-label"] + self.options.get("class-label", []),
            )
            if "no-sync" not in self.options:
                tab_label["sync_id"] = f"tabcode-{language}"
            tab_content = create_component(
                "tab-content",
                children=[item],
                classes=["sd-tab-content"] + self.options.get("class-content", []),
            )
            tab_item = create_component(
                "tab-item",
                children=[tab_label, tab_content],
                classes=["sd-tab-item"] + self.options.get("class-item", []),
            )
            new_children.append(tab_item)
        tab_set.children = new_children
        return [tab_set]


class sd_tab_input(nodes.Element, nodes.General):
    pass


class sd_tab_label(nodes.TextElement, nodes.General):
    pass


def visit_tab_input(self, node):
    attributes = {"ids": [node["id"]], "type": node["type"], "name": node["set_id"]}
    if node["checked"]:
        attributes["checked"] = "checked"
    self.body.append(self.starttag(node, "input", **attributes))


def depart_tab_input(self, node):
    self.body.append("</input>")


def visit_tab_label(self, node):
    attributes = {"for": node["input_id"]}
    if "sync_id" in node:
        attributes["data-sync-id"] = node["sync_id"]
    self.body.append(self.starttag(node, "label", **attributes))


def depart_tab_label(self, node):
    self.body.append("</label>")


class TabSetHtmlTransform(SphinxPostTransform):
    """Transform tab-set to HTML specific AST structure."""

    default_priority = 200
    formats = ("html",)

    def run(self) -> None:
        """Run the transform."""

        # setup id generators
        tab_set_id_base = "sd-tab-set-"
        tab_set_id_num = 0
        tab_item_id_base = "sd-tab-item-"
        tab_item_id_num = 0

        for tab_set in findall(self.document)(
            lambda node: is_component(node, "tab-set")
        ):
            tab_set_identity = tab_set_id_base + str(tab_set_id_num)
            tab_set_id_num += 1
            children = []
            # get the first selected node
            selected_idx = None
            for idx, tab_item in enumerate(tab_set.children):
                if tab_item.get("selected", False):
                    if selected_idx is None:
                        selected_idx = idx
                    else:
                        LOGGER.warning(
                            f"Multiple selected 'tab-item' directives [{WARNING_TYPE}.tab]",
                            location=tab_item,
                            type=WARNING_TYPE,
                            subtype="tab",
                        )
            selected_idx = 0 if selected_idx is None else selected_idx

            for idx, tab_item in enumerate(tab_set.children):
                try:
                    tab_label, tab_content = tab_item.children
                except ValueError:
                    print(tab_item)
                    raise
                tab_item_identity = tab_item_id_base + str(tab_item_id_num)
                tab_item_id_num += 1

                # create: <input checked="checked" id="id" type="radio">
                input_node = sd_tab_input(
                    "",
                    id=tab_item_identity,
                    set_id=tab_set_identity,
                    type="radio",
                    checked=(idx == selected_idx),
                )
                input_node.source, input_node.line = tab_item.source, tab_item.line
                children.append(input_node)

                # create: <label for="id">...</label>
                label_node = sd_tab_label(
                    "",
                    *tab_label.children,
                    input_id=tab_item_identity,
                    classes=tab_label["classes"],
                )
                if tab_label.get("ids"):
                    label_node["ids"] += tab_label["ids"]
                if "sync_id" in tab_label:
                    label_node["sync_id"] = tab_label["sync_id"]
                label_node.source, label_node.line = tab_item.source, tab_item.line
                children.append(label_node)

                # add content
                children.append(tab_content)

            tab_set.children = children
